from types import SimpleNamespace
from typing import Union, Literal
from unrealircd_rpc_py.Connection import Connection
import unrealircd_rpc_py.Definition as Dfn

class Channel:

    DB_CHANNELS: list[Dfn.Channel] = []

    def __init__(self, connection: Connection) -> None:

        # Store the original response
        self.response_raw: str = ''
        """Original response used to see available keys."""

        self.response_np: SimpleNamespace
        """Parsed JSON response providing access to all keys as attributes."""

        # Get the Connection instance
        self.Connection = connection
        self.Logs = connection.Logs
        self.Error = connection.Error

    @property
    def get_error(self) -> Dfn.RPCError:
        return self.Error

    @property
    def get_response(self) -> Union[dict, None]:
        return self.Connection.get_response()

    @property
    def get_response_np(self) -> Union[SimpleNamespace, None]:
        return self.Connection.get_response_np()

    def list_(self, object_detail_level: Literal[0, 1, 2, 3, 4] = 1) -> list[Dfn.Channel]:
        """List channels.

        if you want to have more details increase the level or see the level you want by visiting this page:

        https://www.unrealircd.org/docs/JSON-RPC:Channel#Structure_of_a_channel

        Args:
            object_detail_level (int, optional): set the detail of the response object, see the Detail level column in Structure of a channel. In this RPC call it defaults to 1 if this parameter is not specified. Defaults to 1.

        Returns:
            Channel: List of Channel object, None if nothing see the Error property
        """
        try:
            self.DB_CHANNELS = []
            self.Connection.EngineError.init_error()

            response:dict[str, dict] = self.Connection.query(method='channel.list', param={'object_detail_level': object_detail_level})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return self.DB_CHANNELS

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return self.DB_CHANNELS

            channels: list[dict] = response['result']['list']

            for channel in channels:
                channel_copy: dict = channel.copy()

                for key in ['bans','ban_exemptions','invite_exceptions', 'members']:
                    channel_copy.pop(key, None)

                self.DB_CHANNELS.append(
                        Dfn.Channel(
                            **channel_copy,
                            bans=[Dfn.ChannelBans(**ban) for ban in channel.get('bans', [])],
                            ban_exemptions=[Dfn.ChannelBanExemptions(**ban_ex) for ban_ex in channel.get('ban_exemptions', [])],
                            invite_exceptions=[Dfn.ChannelInviteExceptions(**inv_ex) for inv_ex in channel.get('invite_exceptions', [])],
                            members=[Dfn.ChannelMembers(**member) for member in channel.get('members', [])]
                        )
                )

            return self.DB_CHANNELS

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return self.DB_CHANNELS
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return self.DB_CHANNELS

    def get(self, channel: str, object_detail_level: int = 3) -> Union[Dfn.Channel, None]:
        """Retrieve all details of a single channel. 
        This returns more information than a channel.list call, see the end of Structure of a channel.

        Args:
            channel (str):  the name of the channel
            object_detail_level (int, optional): set the detail of the response object, see the Detail level column in Structure of a channel. In this RPC call it defaults to 3 if this parameter is not specified. Defaults to 3.

        Returns:
            Channel: The Channel Object, None if nothing see Error property
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query(method='channel.get', param={'channel': channel, 'object_detail_level': object_detail_level})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return None

            channel: dict = response.get('result', {}).get('channel', {})

            channel_copy: dict = channel.copy()
            for key in ['bans','ban_exemptions','invite_exceptions', 'members']:
                channel_copy.pop(key, None)

            members:list[dict] = channel.get('members', [])
            
            db_members: list[Dfn.ChannelMembers] = []

            for member in members:
                user_dict: dict[str, dict] = member.get('user', Dfn.User().to_dict())
                tls_dict: dict[str, dict] = member.get('tls', Dfn.Tls().to_dict())
                geoip_dict: dict[str, dict] = member.get('geoip', Dfn.Geoip().to_dict())

                if 'security-groups' in user_dict:
                    # rename key
                    user_dict['security_groups'] = user_dict.pop('security-groups')
                
                if 'country-code' in geoip_dict:
                    geoip_dict['country_code'] = geoip_dict.pop('country-code')

                for key in ['user', 'tls', 'geoip']:
                    member.pop(key, None)
                
                user_obj = Dfn.User(**user_dict)
                tls_obj = Dfn.Tls(**tls_dict)
                geoip_obj = Dfn.Geoip(**geoip_dict)

                member_obj = Dfn.ChannelMembers(**member, user=user_obj, tls=tls_obj, geoip=geoip_obj)

                db_members.append(member_obj)

            channel_obj = Dfn.Channel(
                        **channel_copy,
                        bans=[Dfn.ChannelBans(**ban) for ban in channel.get('bans', [])],
                        ban_exemptions=[Dfn.ChannelBanExemptions(**ban_ex) for ban_ex in channel.get('ban_exemptions', [])],
                        invite_exceptions=[Dfn.ChannelInviteExceptions(**inv_ex) for inv_ex in channel.get('invite_exceptions', [])],
                        # members=[Dfn.ChannelMembers(**member) for member in channel.get('members', [])]
                        members=db_members
                    )
            
            return channel_obj

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return None
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return None

    def set_mode(self, channel: str, modes: str, parameters: str = "") -> bool:
        """Set and unset modes on a channel.

        Args:
            channel (str): The name of the channel
            modes (str): The mode(s) to change, eg +be
            parameters (str): The parameters, eg some!nice@ban some!nice@invex

        Returns:
            bool: True if success
        """
        try:
            self.Connection.EngineError.init_error()

            response:dict[str, dict] = self.Connection.query(method='channel.set_mode', param={"channel": channel,"modes": modes,"parameters": parameters})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False

    def set_topic(self, channel: str, topic: str, set_by: str = None, set_at: str = None) -> bool:
        """Set a topic on a channel.

        Args:
            channel (str): The name of the channel
            topic (str): The new topic
            set_by (str, optional): who set the topic (but, see important note below). Defaults to None.
            set_at (str, optional): when the topic was set (timestamp). Defaults to None.

        Returns:
            bool: True if success
        """
        try:
            self.Connection.EngineError.init_error()

            response:dict[str, dict] = self.Connection.query(method='channel.set_topic', param={"channel": channel, "topic": topic, "set_by": set_by, "set_at": set_at})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False

    def kick(self, channel: str, nick: str, reason: str) -> bool:
        """Kick a user from a channel

        Args:
            channel (str): The name of the channel
            nick (str): The user to kick
            reason (str): The kick reason (shown in the channel)

        Returns:
            bool: True if success
        """
        try:
            self.Connection.EngineError.init_error()

            response:dict[str, dict] = self.Connection.query(method='channel.kick', param={"channel": channel, "nick": nick, "reason": reason})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False


from types import SimpleNamespace
from typing import Union
from unrealircd_rpc_py.Connection import Connection
import unrealircd_rpc_py.Definition as Dfn

class Spamfilter:

    DB_SPAMFILTERS: list[Dfn.Spamfilter] = []

    def __init__(self, connection: Connection) -> None:

        # Get the Connection instance
        self.Connection = connection
        self.Logs = connection.Logs
        self.Error = connection.Error

    @property
    def get_error(self) -> Dfn.RPCError:
        return self.Error

    @property
    def get_response(self) -> Union[dict, None]:
        return self.Connection.get_response()

    @property
    def get_response_np(self) -> Union[SimpleNamespace, None]:
        return self.Connection.get_response_np()

    def list_(self) -> list[Dfn.Spamfilter]:
        """List spamfilters.

        Returns:
            list[Spamfilter]: List of Spamfilter, None if nothing, see Error property if any error
        """
        try:
            self.Connection.EngineError.init_error()
            self.DB_SPAMFILTERS = []

            response: dict[str, dict] = self.Connection.query(method='spamfilter.list')

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return self.DB_SPAMFILTERS

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return self.DB_SPAMFILTERS

            spamfilters = response['result']['list']

            for spamfilter in spamfilters:
                self.DB_SPAMFILTERS.append(Dfn.Spamfilter(**spamfilter))

            return self.DB_SPAMFILTERS

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return []
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return []

    def get(self, name: str, match_type: str, ban_action: str, spamfilter_targets: str) -> Union[Dfn.Spamfilter, None]:
        """Retrieve all details of a single spamfilter.

        Mandatory arguments (see structure of a spamfilter for an explanation of the fields):

        https://www.unrealircd.org/docs/JSON-RPC:Spamfilter#Structure_of_a_spamfilter

        Args:
            name (str): The target of the ban or except. For Spamfilter this is the regex/matcher.
            match_type (str): Only for spamfilters! The matching type. One of: simple, regex
            ban_action (str): Only for spamfilters! The action to take on spamfilter hit.
            spamfilter_targets (str): Only for spamfilters! Which targets the spamfilter must filter on.

        Returns:
            Spamfilter: The Object Spamfilter. Could be empty if error
        """
        try:
            self.Connection.EngineError.init_error()

            response:dict[str, dict] = self.Connection.query(method='spamfilter.get', param={"name": name, "match_type": match_type, "ban_action": ban_action, "spamfilter_targets": spamfilter_targets})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return None

            spamfilter = response['result']['tkl']

            object_spamfilter = Dfn.Spamfilter(**spamfilter)

            return object_spamfilter

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return None
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return None

    def add(self, name: str, match_type: str, ban_action: str, ban_duration: int, spamfilter_targets: str, reason: str, set_by: str = None) -> bool:
        """Add a spamfilter.

        Mandatory arguments (see structure of a spamfilter for an explanation of the fields):

        {"name":"regex123","match_type": "regex","ban_action": "gline","ban_duration": 30,"spamfilter_targets": "cpnNPq","reason": "RPC test"}

        https://www.unrealircd.org/docs/JSON-RPC:Spamfilter#Structure_of_a_spamfilter

        Args:
            name (str): The target of the ban or except. For Spamfilter this is the regex/matcher.
            match_type (str): Only for spamfilters! The matching type. One of: simple, regex
            ban_action (str): Only for spamfilters! The action to take on spamfilter hit.
            ban_duration (int): The duration of the ban
            spamfilter_targets (str): Only for spamfilters! Which targets the spamfilter must filter on.
            reason (str): The reason of the ban
            set_by (str, optional): Name of the person or server who set the ban. Default to None

        Returns:
            bool: True if success
        """
        try:
            self.Connection.EngineError.init_error()

            response:dict[str, dict] = self.Connection.query(
                method='spamfilter.add', 
                param={"name": name, "match_type": match_type, "ban_action": ban_action, "ban_duration": ban_duration, "spamfilter_targets": spamfilter_targets, 'reason': reason, 'set_by': set_by}
                )

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False

    def del_(self,  name: str, match_type: str, ban_action: str, spamfilter_targets: str, _set_by: str = None) -> bool:
        """Delete a spamfilter.

        Mandatory arguments (see structure of a spamfilter for an explanation of the fields):

        https://www.unrealircd.org/docs/JSON-RPC:Spamfilter#Structure_of_a_spamfilter

        Args:
            name (str): The target of the ban or except. For Spamfilter this is the regex/matcher.
            match_type (str): Only for spamfilters! The matching type. One of: simple, regex
            ban_action (str): Only for spamfilters! The action to take on spamfilter hit.
            spamfilter_targets (str): Only for spamfilters! Which targets the spamfilter must filter on.
            _set_by (str, optional): Name of the person or server who set the ban. Default to None

        Returns:
            bool: True if success
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query(
                method='spamfilter.del',
                param={"name": name, "match_type": match_type, "ban_action": ban_action, "spamfilter_targets": spamfilter_targets, 'set_by': _set_by}
                )

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False


from types import SimpleNamespace
from typing import Union
from unrealircd_rpc_py.Connection import Connection
import unrealircd_rpc_py.Definition as Dfn

class ServerBan:

    DB_SERVERS_BANS: list[Dfn.ServerBan] = []

    def __init__(self, connection: Connection) -> None:

        # Get the Connection instance
        self.Connection = connection
        self.Logs = connection.Logs
        self.Error = connection.Error

    @property
    def get_error(self) -> Dfn.RPCError:
        return self.Error

    @property
    def get_response(self) -> Union[dict, None]:
        return self.Connection.get_response()

    @property
    def get_response_np(self) -> Union[SimpleNamespace, None]:
        return self.Connection.get_response_np()

    def list_(self) -> list[Dfn.ServerBan]:
        """List server bans (LINEs).

        Returns:
            list[ServerBan]: List of ServerBan, if empty see Error Object
        """
        try:
            self.Connection.EngineError.init_error()
            self.DB_SERVERS_BANS = []

            response:dict[str, dict] = self.Connection.query(method='server_ban.list')

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return self.DB_SERVERS_BANS

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return self.DB_SERVERS_BANS

            srvbans:list[dict] = response.get('result', {}).get('list', [])

            for srvban in srvbans:
                self.DB_SERVERS_BANS.append(Dfn.ServerBan(**srvban))

            return self.DB_SERVERS_BANS

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return []
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return []

    def get(self, query_type: str, name: str) -> Union[Dfn.ServerBan, None]:
        """Retrieve all details of a single server ban (LINE).

        Args:
            query_type (str): Type of the server ban. One of: gline, kline, gzline, zline, spamfilter, qline, except, shun, local-qline, local-exception, local-spamfilter.
            name (str): The target of the ban or except. For Spamfilter this is the regex/matcher.

        Returns:
            ServerBan (ServerBan): The model Object | If None see the Error Attribut
        """
        try:
            self.Connection.EngineError.init_error()

            response:dict[str, dict] = self.Connection.query(method='server_ban.get', param={'type': query_type, 'name': name})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return None

            srvban: dict = response.get('result', {}).get('tkl', {})

            obj = Dfn.ServerBan(**srvban)

            return obj

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return None
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return None

    def add(self, query_type: str, name: str, reason: str, expire_at: str,
            duration_sting: str, set_by: str = None) -> bool:
        """Add a server ban (LINE).

        Mandatory arguments (see structure of a server ban for an explanation of the fields):

        https://www.unrealircd.org/docs/JSON-RPC:Server_ban#Structure_of_a_server_ban

        Args:
            query_type (str): Type of the server ban. One of: gline, kline, gzline, zline, spamfilter, qline, except, shun, local-qline, local-exception, local-spamfilter.
            name (str): The target of the ban or except. For Spamfilter this is the regex/matcher.
            reason (str): The reason of the ban
            expire_at (str): Date/Time when the server ban will expire. NULL means: never.
            duration_sting (str): How long the ban will last from this point in time (human printable). Uses "permanent" for forever.
            set_by (str, optional): Name of the person or server who set the ban. Defaults to None.

        Returns:
            bool: True if success
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query(method='server_ban.add',
                                             param={"type": query_type, "name": name, "reason": reason,
                                                    "expire_at": expire_at, "duration_string": duration_sting,
                                                    'set_by': set_by}
                                             )

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False

    def del_(self, query_type: str, name: str, set_by: str = None) -> bool:
        """Delete a server ban (LINE).

        Mandatory arguments (see structure of a server ban for an explanation of the fields):

        https://www.unrealircd.org/docs/JSON-RPC:Server_ban#Structure_of_a_server_ban

        Args:
            query_type (str): Type of the server ban. One of: gline, kline, gzline, zline, spamfilter, qline, except, shun, local-qline, local-exception, local-spamfilter.
            name (str): The target of the ban or except. For Spamfilter this is the regex/matcher.
            set_by (str, optional): Name of the person or server who set the ban. Defaults to None.

        Returns:
            bool: True if success
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query(method='server_ban.del',
                                             param={"type": query_type, "name": name, 'set_by': set_by})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False



from types import SimpleNamespace
from typing import Union, Literal
from unrealircd_rpc_py.Connection import Connection
import unrealircd_rpc_py.Definition as Dfn


class User:

    DB_USER: list[Dfn.Client] = []

    def __init__(self, connection: Connection) -> None:

        # Get the Connection instance
        self.Connection = connection
        self.Logs = connection.Logs
        self.Error = connection.Error

    @property
    def get_error(self) -> Dfn.RPCError:
        return self.Error

    @property
    def get_response(self) -> Union[dict, None]:
        return self.Connection.get_response()

    @property
    def get_response_np(self) -> Union[SimpleNamespace, None]:
        return self.Connection.get_response_np()

    def list_(self, object_detail_level: Literal[0, 1, 2, 4] = 2) -> list[Dfn.Client]:
        """List users

        Args:
            object_detail_level (int, optional): set the detail of the response object, see the Detail level column in Structure of a client object. Defaults to 2.

        Returns:
            Client ([Dfn.Client]): List of Client Object
        """
        try:
            self.DB_USER = []
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query('user.list', param={'object_detail_level': object_detail_level})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return self.DB_USER

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return self.DB_USER

            users:list[dict] = response.get('result', {}).get('list', []) # response['result']['list']

            for user in users:
                user_for_client = user.copy()
                user_for_user: dict = user.get('user', {}).copy()

                for key in ['geoip','tls','user']:
                    user_for_client.pop(key, None)

                for key in ['channels','security-groups']:
                    user_for_user.pop(key, None)

                user_model = Dfn.User(
                    **user_for_user,
                    security_groups=user.get('user', {}).get('security-groups', []),
                    channels=[Dfn.UserChannel(**chans) for chans in user.get('user', {}).get('channels', [Dfn.UserChannel().to_dict()])]
                )

                self.DB_USER.append(
                        Dfn.Client(
                            **user_for_client,
                            geoip=Dfn.Geoip(**user.get('geoip', Dfn.Geoip().to_dict())),
                            tls=Dfn.Tls(**user.get('tls', Dfn.Tls().to_dict())),
                            user=user_model
                        )
                )

            return self.DB_USER

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            self.Connection.EngineError.set_error(code=-3,message=ke)
            return []
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            self.Connection.EngineError.set_error(code=-3,message=err)
            return []

    def get(self, nickoruid: str) -> Union[Dfn.Client, None]:
        """Get user information

        Args:
            nickoruid (str): The nickname or uid of the user

        Returns:
            Client (Client): The object Client
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query('user.get', {'nick': nickoruid})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return None

            user:dict[str, dict] = response.get('result', {}).get('client', {}) # response['result']['client']

            user_for_client = user.copy()
            user_for_user: dict = user.get('user', {}).copy()

            for key in ['geoip','tls','user']:
                user_for_client.pop(key, None)

            for key in ['channels','security-groups']:
                user_for_user.pop(key, None)

            user_model = Dfn.User(
                **user_for_user,
                security_groups=user.get('user', {}).get('security-groups', []),
                channels=[Dfn.UserChannel(**chans) for chans in user.get('user', {}).get('channels', [])]
            )

            user_obj = Dfn.Client(
                        **user_for_client,
                        geoip=Dfn.Geoip(**user.get('geoip', {})),
                        tls=Dfn.Tls(**user.get('tls', {})),
                        user=user_model
                    )

            return user_obj

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return None
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return None

    def set_nick(self, nickoruid: str, newnick: str, force: bool = False) -> bool:
        """Sets the nick name of a user (changes the nick).
        force: if set to true then q-lines (banned nick) checks will be bypassed. And also, 
        if the new nick name already exists, the other existing user is killed and we will take that new nick.

        Args:
            nickoruid (str): the nick name or the UID
            newnick (str): the new nick name
            force (bool, optional): q-lines (banned nick) checks will be bypassed. Defaults to False.

        Returns:
            bool: True if success else error will be stored in ErrorModel
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query('user.set_nick', {'nick': nickoruid, 'newnick': newnick, 'force': force})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False

    def set_username(self, nickoruid: str, username:str) -> bool:
        """Set the username / ident of a user.

        Args:
            nickoruid (str): the nick name or the UID
            username (str): the new user name / ident

        Returns:
            bool: True if success else error will be stored in ErrorModel
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query('user.set_username', {'nick': nickoruid, 'username': username})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False

    def set_realname(self, nickoruid: str, realname:str) -> bool:
        """Set the realname / gecos of a user.

        Args:
            nickoruid (str): the nick name or the UID
            realname (str):  the new realname / gecos of a user.

        Returns:
            bool: True if success else error will be stored in ErrorModel
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query('user.set_realname', {'nick': nickoruid, 'realname': realname})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False

    def set_vhost(self, nickoruid: str, vhost:str) -> bool:
        """Set a virtual host (vhost) on the user.

        Args:
            nickoruid (str): the nick name or the UID
            vhost (str): the new virtual host.

        Returns:
            bool: True if success else error will be stored in ErrorModel
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query('user.set_vhost', {'nick': nickoruid, 'vhost': vhost})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False

    def set_mode(self, nickoruid: str, modes:str) -> bool:
        """Change the modes of a user.

        Args:
            nickoruid (str): the nick name or the UID
            modes (str): the mode string, eg -i+w

        Returns:
            bool: True if success else error will be stored in ErrorModel
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query('user.set_mode', {'nick': nickoruid, 'modes': modes})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False

    def set_snomask(self, nickoruid: str, snomask:str) -> bool:
        """Change the snomask of a user.

        Args:
            nickoruid (str): the nick name or the UID
            snomask (str): the snomask string, eg -j+R

        Returns:
            bool: True if success else error will be stored in ErrorModel
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query('user.set_snomask', {'nick': nickoruid, 'snomask': snomask})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False

    def set_oper(self, nickoruid: str, oper_account:str, oper_class: str,
                 class_: str = '', modes: str = '',
                 snomask: str = '', vhost: str = ''
                 ) -> bool:
        """Make user an IRC operator.

        Args:
            nickoruid (str): the nick name or the UID
            oper_account (str):  the oper account, to be shown in WHOIS to fellow ircops and in logs.
            oper_class (str):  the operclass. Usually one of the default operclasses like netadmin-with-override
            class_ (str, optional): the class to put the user in. If this option is not specified then opers is assumed, since this class exists in most unrealircd.conf's. You can specify "" (empty) if you don't want to put the user in a class. Defaults to ''.
            modes (str, optional): user modes to set on oper. For example: +xws. If this option is not specified then set::modes-on-oper is used. You can specify "" (empty) if you don't want to set any additional modes on the user. Defaults to ''.
            snomask (str, optional): snomask to set on oper. For example: +bBcdfkqsSoO. If this option is not specified then set::snomask-on-oper is used. You can specify "" (empty) if you don't want to set any snomasks on the user. Defaults to ''.
            vhost (str, optional): virtual host to set on oper. Defaults to ''.

        Returns:
            bool: True if success else error will be stored in ErrorModel
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query('user.set_oper', {'nick': nickoruid, 'oper_account': oper_account, 'oper_class': oper_class, 'class': class_, 'modes': modes, 'snomask': snomask, 'vhost': vhost})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False

    def join(self, nickoruid: str, channel:str, key: str = '', force: bool = False) -> bool:
        """Join a user to a channel.

        Note: If force is set to true then the user will walk through bans, modes and other restrictions 
        (similar to SAJOIN on IRC). If force is not set, or set to false, then it will be a regular JOIN attempt that may fail. 
        If it fails then the user will see an error message on their screen, such as for example Cannot join channel (+k).

        Args:
            nickoruid (str): the nick name or the UID
            channel (str): the channel(s) to join (e.g. #channel or #channel1,#channel2)
            key (str, optional): the key of the channel(s) (only for channels with +k needed, again separate by colon for multiple channels). Defaults to ''.
            force (bool, optional): whether to bypass join restrictions or not. Defaults to False.

        Returns:
            bool: True if success else error will be stored in ErrorModel
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query('user.join', {'nick': nickoruid, 'channel': channel, 'key': key, 'force': force})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False

    def part(self, nickoruid: str, channel:str, force: bool = False) -> bool:
        """Part a user from a channel.

        Note: If force is set to true then the user will see a notice that they were forcefully PARTed from the channel(s). 
        If force is not set, or set to false, then there will be no such notice.

        Args:
            nickoruid (str): the nick name or the UID
            channel (str): the channel(s) to join (e.g. #channel or #channel1,#channel2)
            force (bool, optional): whether to bypass join restrictions or not. Defaults to False.

        Returns:
            bool: True if success else error will be stored in ErrorModel
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query('user.part', {'nick': nickoruid, 'channel': channel, 'force': force})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False

    def kill(self, nickoruid: str, reason:str) -> bool:
        """Kill a user, showing that the user was forcefully removed.

        Note: There is also user.quit. The difference is that user.kill will show that the user is forcefully killed, while user.quit pretend the user quit normally. 
        This slight difference may invoke certain client behavior. For example, some clients don't immediately reconnect for a KILL but do so on a QUIT.

        Args:
            nickoruid (str): the nick name or the UID
            reason (str): reason for the kill

        Returns:
            bool: True if success else error will be stored in ErrorModel
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query('user.kill', {'nick': nickoruid, 'reason': reason})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False

    def quit(self, nickoruid: str, reason:str) -> bool:
        """Quit a user, pretending it was a normal QUIT.

        Note: There is also user.kill. The difference is that user.kill will show that the user is forcefully killed, 
        while user.quit pretend the user quit normally. This slight difference may invoke certain client behavior. 
        For example, some clients don't immediately reconnect for a KILL but do so on a QUIT.

        Args:
            nickoruid (str): the nick name or the UID
            reason (str): reason for the kill

        Returns:
            bool: True if success else error will be stored in ErrorModel
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query('user.quit', {'nick': nickoruid, 'reason': reason})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False



from types import SimpleNamespace
from typing import Union
from unrealircd_rpc_py.Connection import Connection
import unrealircd_rpc_py.Definition as Dfn

class Stats:

    def __init__(self, connection: Connection) -> None:

        # Get the Connection instance
        self.Connection = connection
        self.Logs = connection.Logs
        self.Error = connection.Error

    @property
    def get_error(self) -> Dfn.RPCError:
        return self.Error

    @property
    def get_response(self) -> Union[dict, None]:
        return self.Connection.get_response()

    @property
    def get_response_np(self) -> Union[SimpleNamespace, None]:
        return self.Connection.get_response_np()

    def get(self, object_detail_level:int = 1) -> Union[Dfn.Stats, None]:
        """Get some quick basic statistics.

        Args:
            object_detail_level (int, optional): set the detail of the response object. The user.countries will be included from level 1+. The default object_detail_level is 1. Defaults to 1.

        Returns:
            list[ModelStats]: The Stats model or None if there is an error see Error Property
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query('stats.get', param={'object_detail_level': object_detail_level})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return None

            result: dict = response.get('result', {})
            
            server_obj = Dfn.StatsServer(**result.get('server', Dfn.StatsServer().to_dict()))
            channel_obj = Dfn.StatsChannel(**result.get('channel', Dfn.StatsChannel().to_dict()))
            server_ban_obj = Dfn.StatsServerBan(**result.get('server_ban', Dfn.StatsServerBan().to_dict()))
            
            user_dict: dict[str, dict] = result.get('user', {})

            # list of stats user counrties
            user_country_list_obj: list[Dfn.StatsUserCountries] = []
            user_countries_list: list[dict] = user_dict.get('countries', [])
            for user_country in user_countries_list:
                user_country_list_obj.append(Dfn.StatsUserCountries(**user_country))

            if 'countries' in user_dict:
                user_dict.pop('countries')

            user_obj = Dfn.StatsUser(**user_dict, countries=user_country_list_obj)

            stats_obj = Dfn.Stats(
                server=server_obj,
                user=user_obj,
                channel=channel_obj,
                server_ban=server_ban_obj
            )

            return stats_obj

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return None
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return None



import requests
import json
import urllib3
import socket
import re
import os
import base64
import ssl
import time
import logging
import random
from requests.auth import HTTPBasicAuth
from typing import Literal, Union, Optional, Any
from types import SimpleNamespace
from unrealircd_rpc_py.EngineError import EngineError

class Connection:

    def __init__(self, req_method:str, url: str, path_to_socket_file: str, username: str, password: str, debug_level: int) -> None:

        self.debug_level = debug_level
        self.Logs: logging.Logger
        self.__init_log_system()

        self.EngineError = EngineError()
        """Engine Error: should be used to set errors"""
        self.Error = self.EngineError.Error
        """Error attribut: to be used to print errors"""

        self.url: Optional[str] = url
        self.path_to_socket_file = path_to_socket_file

        self.endpoint: Optional[str] = None
        self.host: Optional[str] = None
        self.port: int = 0
        self.username = username
        self.password = password

        self.request: str = ''
        self.req_method = req_method

        # Option 2 with Namespacescs
        self.__response: Optional[dict] = {}
        self.__response_np: Optional[SimpleNamespace] = SimpleNamespace()

        self.query('stats.get')

    def __check_url(self, url: str) -> bool:
        """Check provided url if it follow the format

        Args:
            url (str): Url to jsonrpc https://your.rpcjson.link:port/api

        Returns:
            bool: True if url is correct else False
        """
        try:
            response = False

            if url is None:
                return response

            pattern = r'https?://([a-zA-Z0-9\.-]+):(\d+)/(.+)'

            match = re.match(pattern, url)

            if not match is None:
                self.host = match.group(1)
                self.port = match.group(2)
                self.endpoint = match.group(3)
                response = True
            else:
                self.EngineError.set_error(
                    code=-1,
                    message='You must provide the url in this format: https://your.rpcjson.link:port/api'
                )

            return response
        except NameError as nameerr:
            self.Logs.critical(f'NameError: {nameerr}')
            return False

    def __check_unix_socket_file(self, path_to_socket_file: str) -> bool:
        """Check provided full path to socket file if it exist

        Args:
            path_to_socket_file (str): Full path to unix socket file

        Returns:
            bool: True if path is correct else False
        """
        try:
            response = False

            if path_to_socket_file is None:
                self.EngineError.set_error(
                    code=-1,
                    message='The Path to your socket file is empty ? please be sure that you are providing the correct socket path'
                )
                return response

            if not os.path.exists(path_to_socket_file):
                self.EngineError.set_error(
                    code=-1,
                    message='The Path to your socket file is wrong ? please make sure that you are providing the correct socket path'
                )
                return response

            response = True

            return response
        except NameError as nameerr:
            self.Logs.critical(f'NameError: {nameerr}')
            return False

    def __is_error_connection(self, response: str) -> bool:
        """If True, it means that there is an error

        Args:
            response (str): The response to analyse

        Returns:
            bool: True if there is a connection error
        """
        if 'authentication required' == response.lower().strip():
            self.EngineError.set_error(
                code=-1,
                message='>> Authentication required <<'
            )
            return True
        elif 'authentication required' in response.lower().strip():
            self.EngineError.set_error(
                code=-1,
                message='>> Authentication required <<'
            )
            return True
        elif "('Connection aborted.'," in response.lower().strip():
            self.EngineError.set_error(
                code=-1,
                message='>> Connection aborted <<'
            )
            return True
        else:
            return False

    def __send_to_unixsocket(self):
        sock = socket.socket(socket.AddressFamily.AF_UNIX, socket.SocketKind.SOCK_STREAM)
        try:

            if not self.__check_unix_socket_file(self.path_to_socket_file):
                return None

            sock.connect(self.path_to_socket_file)
            sock.settimeout(10)

            if not self.request:
                return None

            sock.sendall(f'{self.request}\r\n'.encode())

            response = b""
            chunk = b""
            pattern = b'\n$'

            while True:
                chunk = sock.recv(4096)
                response += chunk
                if re.findall(pattern, chunk):
                    break

            str_data = response.decode()
            self.__set_responses(str_data)

        except AttributeError as attrerr:
            self.Logs.critical(f'AF_Unix Error: {attrerr}')
            self.EngineError.set_error(code=-1, message=attrerr.__str__())
        except OSError as oserr:
            self.Logs.critical(f'System Error: {oserr}')
            self.EngineError.set_error(code=-1, message=oserr.__str__())
        except Exception as err:
            self.Logs.error(f'General Error: {err}')
            self.EngineError.set_error(code=-1, message=err.__str__())
        finally:
            sock.close()

    def __send_srequest(self):
        """S For socket connection"""
        try:

            if not self.__check_url(self.url) and not self.url is None:
                self.Logs.critical('You must provide the url in this format: https://your.rpcjson.link:port/api')
                return None

            get_host = self.host
            get_port = self.port
            get_username = self.username
            get_password = self.password
            credentials = base64.b64encode(f"{get_username}:{get_password}".encode()).decode()

            # Create a socket and wrap it in an SSL context for HTTPS
            context = ssl.create_default_context()
            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE

            with socket.create_connection((get_host, get_port)) as sock:
                with context.wrap_socket(sock, server_hostname=get_host) as ssock:
                    # Send the HTTPS

                    json_request = self.request
                    request = self.__build_headers(credentials, str(json_request))
                    ssock.sendall(request.encode('utf-8'))

                    # Receive the response
                    response = b""
                    while True:
                        chunk = ssock.recv(4096)
                        if not chunk:
                            break
                        response += chunk

                    # Decode and print the response
                    response_str = response.decode('utf-8')
                    header_end = response_str.find("\r\n\r\n")
                    if header_end != -1:
                        body = response_str[header_end + 4:]  # Extract the body
                    else:
                        body = response_str

                    if self.__is_error_connection(body):
                        return None

                    self.__set_responses(body)

        except (socket.error, ssl.SSLError) as serror:
            self.Logs.error(f'Socket Error: {serror}')
            self.EngineError.set_error(code=-1, message=f'Socket Error: {serror}')
        except json.JSONDecodeError as jdecErr:
            self.Logs.error(f'JSONDecodeError: {jdecErr}')
            self.EngineError.set_error(code=-1, message=f'Json Decode Error: {jdecErr}')
        except Exception as err:
            self.Logs.error(f'General Error: {err}')
            self.EngineError.set_error(code=-1, message=f'General Error: {err}')

    def __send_request(self):
        """Use requests module"""
        try:

            if not self.__check_url(self.url) and not self.url is None:
                self.Logs.critical('You must provide the url in this format: https://your.rpcjson.link:port/api')
                return None

            verify = False
            url: Optional[str] = self.url

            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

            credentials = HTTPBasicAuth(self.username, self.password)
            jsonrequest = self.request

            response = requests.post(url=url, auth=credentials, data=jsonrequest, verify=verify)

            if response.status_code != 200:
                self.EngineError.set_error(code=-1, message=f"Status code {response.status_code} | {response.reason}")
                return None
            
            if self.__is_error_connection(response.text):
                return None

            decoded_response = json.dumps(response.text)
            self.__set_responses(json.loads(decoded_response))

        except KeyError as ke:
            self.Logs.error(f"KeyError : {ke}")
            self.Logs.error(f"Initial request: {self.request}")
            self.EngineError.set_error(code=-1, message=ke.__str__())
        except requests.ReadTimeout as rt:
            self.Logs.error(f"Timeout : {rt}")
            self.Logs.error(f"Initial request: {self.request}")
            self.EngineError.set_error(code=-1, message=f'Timeout: {rt}')
        except requests.ConnectionError as ce:
            self.Logs.error(f"Connection Error : {ce}")
            self.Logs.error(f"Initial request: {self.request}")
            if 'connection aborted.' in str(ce).lower().strip():
                self.EngineError.set_error(code=-1, message=f'>> Connection Aborted <<')
        except json.decoder.JSONDecodeError as jsonerror:
            self.Logs.error(f"jsonError {jsonerror}")
            self.Logs.error(f"Initial request: {self.request}")
            self.EngineError.set_error(code=-1, message=f'Json Error: {jsonerror}')
        except Exception as err:
            self.Logs.error(f"General Error {err}")
            self.Logs.error(f"Initial request: {self.request}")
            self.EngineError.set_error(code=-1, message=f'General Error: {err}')

    def __build_headers(self, credentials: str, data:str) -> str:
        """Build the header for socket connection only

        Args:
            credentials (str): crypted credentials
            data (str): data we need to send to the server

        Returns:
            str: Returd the built header
        """
        headers: str = (
                    f"POST /{self.endpoint} HTTP/1.1\r\n"
                    f"Host: {self.host}\r\n"
                    "Content-Type: application/json\r\n"
                    f"Authorization: Basic {credentials}\r\n"
                    f"Content-Length: {len(data)}\r\n"
                    "Connection: close\r\n"
                    "\r\n"
                    f"{data}"
                )

        return headers

    def __init_log_system(self) -> None:
        """Init log system
        """
        # Init logs object
        self.Logs: logging.Logger = logging.getLogger('unrealircd-rpc-py')
        self.Logs.propagate = False
        self.Logs.setLevel(self.debug_level)

        # Add Handlers
        stdout_hanlder = logging.StreamHandler()
        stdout_hanlder.setLevel(self.debug_level)

        # Define log format
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(filename)s - %(lineno)d - %(funcName)s - %(message)s')

        # Apply log format
        stdout_hanlder.setFormatter(formatter)

        # Add handler to logs
        self.Logs.addHandler(stdout_hanlder)

    def query(self,
              method: Union[Literal['stats.get', 'rpc.info','user.list'], str],
              param: Optional[dict] = None,
              query_id: int = 123,
              jsonrpc: str = '2.0'
              ) -> Union[dict, None]:
        """This method will use to run the queries

        Args:
            method (Union[Literal[&#39;stats.get&#39;, &#39;rpc.info&#39;,&#39;user.list&#39;], str]): The method to send to unrealircd
            param (dict, optional): the paramaters to send to unrealircd. Defaults to {}.
            query_id (int, optional): id of the request. Defaults to 123.
            jsonrpc (str, optional): jsonrpc. Defaults to '2.0'.

        Returns:
            str: The correct response
            None: no response from the server
            bool: False if there is an error occured
        """

        # data = '{"jsonrpc": "2.0", "method": "stats.get", "params": {}, "id": 123}'
        get_method = method
        get_param = {} if param is None else param

        if query_id == 123:
            rand = random.randint(1, 6000)
            get_id = int(time.time()) + rand

        else:
            get_id = query_id

        response = {
            "jsonrpc": jsonrpc,
            "method": get_method,
            "params": get_param,
            "id": get_id
        }

        self.request = json.dumps(response)

        if self.req_method == 'socket':
            self.__send_srequest()
        elif self.req_method == 'unixsocket':
            self.__send_to_unixsocket()
        elif self.req_method == 'requests':
            self.__send_request()
        else:
            self.Logs.warning('No valid request method')
            self.EngineError.set_error(code=-1, message='<< Invalid method >>')

        if self.__response == '' or self.__response is None:
            return None

        return self.get_response()

    def get_keys_levels(self, data: Any, prefix=''):
        """Parse the Json output and list all available keys
        """

        if isinstance(data, dict):
            for key, value in data.items():
                complete_key = f"{prefix}.{key}" if prefix else key
                print(f"Key : {complete_key}, Type : {type(value).__name__}")
                self.get_keys_levels(value, complete_key)

        elif isinstance(data, list):
            for index, element in enumerate(data):
                complete_key = f"{prefix}[{index}]"
                print(f"Key : {complete_key}, Type : list")
                self.get_keys_levels(element, complete_key)

    def dict_to_namespace(self, dictionary: dict):

        if isinstance(dictionary, dict):
            return SimpleNamespace(**{key: self.dict_to_namespace(value) for key, value in dictionary.items()})

        return dictionary

    def __set_responses(self, response: str):
        """Set response as dict and as simple name space"""
        # Set dict response
        self.__response_np: Optional[SimpleNamespace] = None
        self.__response: Optional[dict] = None

        if not response:
            self.Logs.error(f"Impossible to load response: {response}")
            return

        self.__response = json.loads(response)

        if not isinstance(self.__response, dict):
            self.__response = None
            self.__response_np = None
            self.Logs.error(f"Impossible to load response: {response}")

        # Set response name space
        self.__response_np = self.dict_to_namespace(self.__response)

    def get_response_np(self) -> Union[SimpleNamespace, None]:
        return self.__response_np

    def get_response(self) -> Union[dict, None]:
        return self.__response


from dataclasses import dataclass, field, asdict, fields
from json import dumps
from typing import Any, Optional
from warnings import warn
from functools import wraps

def deprecated(reason: str = "Deprecated"):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            warn(
                f"{func.__name__}() deprecated : {reason}",
                category=DeprecationWarning,
                stacklevel=2
            )
            return func(*args, **kwargs)
        return wrapper
    return decorator

@dataclass
class MainModel:
    """Parent Model contains important methods"""
    def to_dict(self) -> dict[str, Any]:
        """Return the fields of a dataclass instance as a new dictionary mapping field names to field values."""
        return asdict(self)
    
    def to_json(self) -> str:
        """Return the object of a dataclass a json str."""
        return dumps(self.to_dict())

    def get_attributes(self) -> list[str]:
        """Return a list of attributes name"""
        return [f.name for f in fields(self)]


@dataclass
class RPCError(MainModel):
    """This model will contain the error if any"""
    code: int = 0
    message: Optional[str] = None

@dataclass
class LiveRPCError(MainModel):
    """This the Live JSONRPC Model Error"""
    jsonrpc: str = "2.0"
    error: RPCError = field(default_factory=RPCError())
    id: int = 123

@dataclass
class LiveRPCResult(MainModel):
    """This the Live JSONRPC Model Result"""
    jsonrpc: str = "2.0"
    method: Optional[str] = None
    result: Optional[Any] = None
    id: int = 123

@dataclass
class Tls(MainModel):
    certfp: str = None
    cipher: str = None

@dataclass
class Geoip(MainModel):
    country_code: str = None
    asn: str = None
    asname: str = None

#################
#   User Class  #
#################

@dataclass
class UserChannel(MainModel):
    """User Class
    """
    name: str = None
    level: str = None

@dataclass
class User(MainModel):
    """User Class
    
    Depends on:
        ```
        1- UserChannel
        ```
    """
    username: str = None
    realname: str = None
    vhost: str = None
    cloakedhost: str = None
    servername: str = None
    account: str = None
    reputation: int = None
    security_groups: list[str] = field(default_factory=list)
    modes: str = None
    snomasks: str = None
    operlogin: str = None
    operclass: str = None
    channels: list[UserChannel] = field(default_factory=list[UserChannel])

@dataclass
class Client(MainModel):
    """User Class
    
    Depends on:
        ```
        1- Geoip
        2- Tls
        3- User
        ```
    """
    name: str = None
    id: str = None
    hostname: str = None
    ip: str = None
    details: str = None
    geoip: Geoip = field(default_factory=Geoip)
    server_port: int = 0
    client_port: int = 0
    connected_since: str = None
    idle_since: str = None
    tls: Tls = field(default_factory=Tls)
    user: User = field(default_factory=User)


#################
# Server Class #
#################

@dataclass
class ServerModule(MainModel):
    """Server Class (.module_list)
    """
    name: str = None
    version: str = None
    author: str = None
    description: str = None
    third_party: bool = False
    permanent: bool = False
    permanent_but_reloadable: bool = False

@dataclass
class ServerRehashClient(MainModel):
    """Server Class (.rehash)
    """
    name: str = None
    id: str = None
    hostname: str = None
    ip: str = None
    server_port: int = 0
    details: str = None
    connected_since: str = None
    idle_since: str = None

@dataclass
class ServerRehashLogSource(MainModel):
    """Server Class (.rehash)
    """
    file: str = None
    line: int = 0
    function: str = None

@dataclass
class ServerRehashLog(MainModel):
    """Server Class (.rehash)
    
    Depends on:
        ```
        1- ServerRehashLogSource
        ```
    """
    timestamp: str = None
    level: str = None
    subsystem: str = None
    event_id: str = None
    log_source: str = None
    msg: str = None
    source: ServerRehashLogSource = field(default_factory=ServerRehashLogSource)

@dataclass
class ServerRehash(MainModel):
    """Server Class (.rehash)
    
    Depends on:
        ```
        1- ServerRehashClient
        2- ServerRehashLog
        ```
    """
    rehash_client: ServerRehashClient = field(default_factory=ServerRehashClient)
    log: list[ServerRehashLog] = field(default_factory=list[ServerRehashLog])
    success: str = None

@dataclass
class ServerRpcModules(MainModel):
    name: str = None
    version: str = None

@dataclass
class ServerFeatures(MainModel):
    """Server Class
    
    Depends on:
        ```
        1- ServerRpcModules
        ```
    """
    software: str = None
    protocol: int = 0
    usermodes: str = None
    chanmodes: list[str] = field(default_factory=list)
    nick_character_sets: str = None
    rpc_modules: list[ServerRpcModules] = field(default_factory=list[ServerRpcModules])

@dataclass
class Server(MainModel):
    """Server Class
    
    Depends on:
        ```
        1- ServerFeatures
        ```
    """
    info: str = None
    uplink: str = None
    num_users: int = 0
    boot_time: str = None
    synced: bool = False
    ulined: bool = False
    features: ServerFeatures = field(default_factory=ServerFeatures)

@dataclass
class ClientServer(MainModel):
    """Server Class
    
    Depends on:
        ```
        1- Server
        2- Tls
        ```
    """
    name: str = None
    id: str = None
    hostname: str = None
    ip: str = None
    details: str = None
    server_port: int = 0
    client_port: int = 0
    connected_since: str = None
    idle_since: str = None
    server: Server = field(default_factory=Server)
    tls: Tls = field(default_factory=Tls)

#################
# Channel Class #
#################
@dataclass
class ChannelBans(MainModel):
    """Channel Class 
    """
    name: str = None
    set_by: str = None
    set_at: str = None

@dataclass
class ChannelBanExemptions(MainModel):
    """Channel Class 
    """
    name: str = None
    set_by: str = None
    set_at: str = None

@dataclass
class ChannelInviteExceptions(MainModel):
    """Channel Class 
    """
    name: str = None
    set_by: str = None
    set_at: str = None

@dataclass
class ChannelMembers(MainModel):
    """Channel Class 
    
    Depends on:
        ```
        1- Geoip
        ```
    """
    level: str = None
    name: str = None
    id: str = None
    hostname: str = None
    ip: str = None
    details: str = None
    server_port: int = 0
    client_port: int = 0
    connected_since: str = None
    idle_since: str = None
    user: User = field(default_factory=User)
    geoip: Geoip = field(default_factory=Geoip)
    tls: Tls = field(default_factory=Tls)

@dataclass
class Channel(MainModel):
    """Channel Class 
    
    Depends on:
        ```
        1- ChannelBans
        2- ChannelBanExemptions
        3- ChannelInviteExceptions
        4- ChannelMembers
        ```
    """
    name: str = None
    creation_time: str = None
    num_users: int = 0
    topic: str = None
    topic_set_by: str = None
    topic_set_at: str = None
    modes: str = None
    bans: list[ChannelBans] = field(default_factory=list[ChannelBans])
    ban_exemptions: list[ChannelBanExemptions] = field(default_factory=list[ChannelBanExemptions])
    invite_exceptions: list[ChannelInviteExceptions] = field(default_factory=list[ChannelInviteExceptions])
    members: list[ChannelMembers] = field(default_factory=list[ChannelMembers])

@dataclass
class NameBan(MainModel):
    """Name Ban class"""
    type: str = None
    type_string: str = None
    set_by: str = None
    set_at: str = None
    expire_at: str = None
    set_at_string: str = None
    expire_at_string: str = None
    duration_string: str = None
    set_at_delta: int = 0
    set_in_config: bool = False
    name: str = None
    reason: str = None

@dataclass
class ServerBan(MainModel):
    """Server Ban class"""
    type: str = None
    type_string: str = None
    set_by: str = None
    set_at: str = None
    expire_at: str = None
    set_at_string: str = None
    expire_at_string: str = None
    duration_string: str = None
    set_at_delta: int = 0
    set_in_config: bool = False
    name: str = None
    reason: str = None

@dataclass
class ServerBanException(MainModel):
    """Server Ban Exception class"""
    type: str = None
    type_string: str = None
    set_by: str = None
    set_at: str = None
    expire_at: str = None
    set_at_string: str = None
    expire_at_string: str = None
    duration_string: str = None
    set_at_delta: int = None
    set_in_config: bool = False
    name: str = None
    reason: str = None
    exception_types: str = None

@dataclass
class Spamfilter(MainModel):
    """Spamfilters class"""
    type: str = None
    type_string: str = None
    set_by: str = None
    set_at: str = None
    expire_at: str = None
    set_at_string: str = None
    expire_at_string: str = None
    duration_string: str = None
    set_at_delta: int = 0
    set_in_config: bool = False
    name: str = None
    match_type: str = None
    ban_action: str = None
    ban_duration: int = 0
    ban_duration_string: str = None
    spamfilter_targets: str = None
    reason: str = None
    hits: int = 0
    hits_except: int = 0

@dataclass
class RpcInfo(MainModel):
    """Rpc Class"""
    name: str = None
    module: str = None
    version: str = None

#################
#  Stats Class  #
#################

@dataclass
class StatsServer(MainModel):
    """Stats Class
    """
    total: int = 0
    ulined: int = 0

@dataclass
class StatsUserCountries(MainModel):
    """Stats Class
    """
    country: str = None
    count: int = 0

@dataclass
class StatsUser(MainModel):
    """Stats Class 
    
    Depends on:
        ```
        1- StatsUserCountries
        ```
    """
    total: int = 0
    ulined: int = 0
    oper: int = 0
    record: int = 0
    countries: list[StatsUserCountries] = field(default_factory=list[StatsUserCountries])

@dataclass
class StatsServerBan(MainModel):
    """Stats Class
    """
    total: int = 0
    server_ban: int = 0
    spamfilter: int = 0
    name_ban: int = 0
    server_ban_exception: int = 0

@dataclass
class StatsChannel(MainModel):
    """Stats Class
    """
    total: int = 0

@dataclass
class Stats(MainModel):
    """Stats Class 
    
    Depends on:
        ```
        1- StatsServer
        2- StatsUser
        3- StatsChannel
        4- StatsServerBan
        ```
    """
    server: StatsServer = field(default_factory=StatsServer)
    user: StatsUser = field(default_factory=StatsUser)
    channel: StatsChannel = field(default_factory=StatsChannel)
    server_ban: StatsServerBan = field(default_factory=StatsServerBan)

##################
#  Whowas Class  #
##################

@dataclass
class WhowasUser(MainModel):
    username: str = None
    realname: str = None
    vhost: str = None
    servername: str = None
    account: str = None

@dataclass
class Whowas(MainModel):
    """Whowas Class 
    
    Depends on:
    ```
    1- WhowasUser
    2- Geoip
    ```
    """
    name: str = None
    event: str = None
    logon_time: str = None
    logoff_time: str = None
    hostname: str = None
    ip: str = None
    details: str = None
    connected_since: str = None
    user: WhowasUser = field(default_factory=WhowasUser)
    geoip: Geoip = field(default_factory=Geoip)


from types import SimpleNamespace
from typing import Union
from unrealircd_rpc_py.Connection import Connection
import unrealircd_rpc_py.Definition as Dfn

class Rpc:

    DB_RPC_INFO: list[Dfn.RpcInfo] = []

    def __init__(self, connection: Connection) -> None:

        # Store the original response
        self.response_raw: str
        """Original response used to see available keys."""

        self.response_np: SimpleNamespace
        """Parsed JSON response providing access to all keys as attributes."""

        # Get the Connection instance
        self.Connection = connection
        self.Logs = connection.Logs
        self.Error = connection.Error

    @property
    def get_error(self) -> Dfn.RPCError:
        return self.Error

    @property
    def get_response(self) -> Union[dict, None]:
        return self.Connection.get_response()

    @property
    def get_response_np(self) -> Union[SimpleNamespace, None]:
        return self.Connection.get_response_np()

    def info(self) -> list[Dfn.RpcInfo]:
        """A response object, with in the result object a "methods" object which is a list of: the API method with in that the name, module name and module version.

        Returns:
            list[RpcInfo]: List of RpcInfo, empty object if error
        """
        try:
            self.DB_RPC_INFO = []
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query(method='rpc.info')

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return []

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return []

            rpcinfos: dict[dict, dict] = response['result']['methods']

            for rpcinfo in rpcinfos:
                self.DB_RPC_INFO.append(
                        Dfn.RpcInfo(**rpcinfos.get(rpcinfo, {}))
                )

            return self.DB_RPC_INFO

        except KeyError as ke:
            self.Logs.error(ke)
            return []
        except Exception as err:
            self.Logs.error(err)
            return []

    def set_issuer(self, name: str) -> bool:
        """Set who is the issuer of all the subsequent commands that are done. 

        This information will be used by the logging system and communicate to servers and IRCOps in the unrealircd.org/issued-by message tag.

        This method only exists in UnrealIRCd 6.1.0 and later

        Args:
            name (str): the name of the person who is issuing the commands, eg. the current logged in person to the admin panel. Note: This name must conform to the regular nickname rules, eg. may not contain spaces and the like.

        Returns:
            bool: True if success
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query(
                method='rpc.set_issuer', 
                param={"name": name}
                )

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return True
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return True

    def add_timer(self, timer_id: str, every_msec: int, request: dict) -> bool:
        """Add a timer so a JSON-RPC request is executed at certain intervals.

        This method only exists in UnrealIRCd 6.1.0 and later

        exemple of request:

        {"jsonrpc": "2.0", "method": "stats.get", "params": {}, "id": 555}

        Args:
            timer_id (str): the name/id of the timer, this so you can cancel it via rpc.del_timer later. It must be unique for this RPC connection.
            every_msec (int): the timer will be executed every -this- milliseconds. The minimum value is 250 (a quarter of a second).
            request (dict): the full JSON-RPC request, so {"jsonrpc":"2.0" etc. etc.

        Returns:
            bool: True if success
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query(
                method='rpc.add_timer',
                param={"timer_id": timer_id, "every_msec": every_msec, "request": request}
                )

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False

    def del_timer(self, timer_id: str) -> bool:
        """Remove a previously added timer. Note that you can only cancel timers that belong to your own connection.

        This method only exists in UnrealIRCd 6.1.0 and later

        Args:
            timer_id (str): the name/id of the timer that needs to be canceled.

        Returns:
            bool: True if success
        """
        try:
            self.Connection.EngineError.init_error()

            response:dict[str, dict] = self.Connection.query(
                method='rpc.del_timer', 
                param={"timer_id": timer_id}
                )

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False



from types import SimpleNamespace
from typing import Union, Literal, Optional
from time import time
from unrealircd_rpc_py.Connection import Connection
from unrealircd_rpc_py.Definition import RPCError


class Log:

    def __init__(self, connection: Connection) -> None:

        # Get the Connection instance
        self.Connection = connection
        self.Logs = connection.Logs
        self.Error = connection.Error

    @property
    def get_error(self) -> RPCError:
        return self.Error

    @property
    def get_response(self) -> Union[dict, None]:
        return self.Connection.get_response()

    @property
    def get_response_np(self) -> Union[SimpleNamespace, None]:
        return self.Connection.get_response_np()

    def list_(self, sources: Optional[list] = None) -> Union[SimpleNamespace, None]:
        """Fetch past log entries (since boot).

        Returns:
            list[ModelNameBan]: List of ModelNameBan, None if nothing or Error see the property
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query(
                method='log.list', 
                param={"sources": sources}
                )

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return None

            return self.get_response_np

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return None
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return None

    def send(self, msg: str, level: Literal['debug','info','warn','error','fatal'],
             subsystem: str, event_id: str, timestamp: float = time()
             ) -> bool:
        """ Send a log message / server notice.

        Requires UnrealIRCd 6.1.8 or later

        Args:
            msg (str): The message you want to send to the server
            level (Literal[&#39;debug&#39;,&#39;info&#39;,&#39;warn&#39;,&#39;error&#39;,&#39;fatal&#39;]): You can pick the level in the list
            subsystem (str): You can put anything you want as a subsystem
            event_id (str): You can put anything you want as a event_id
            timestamp (float, optional): The current timestamp. Defaults to time().

        Returns:
            bool: True if is okey, else if there is an error
        """
        # waiting for the documentation.
        try:
            self.Connection.EngineError.init_error()

            response = self.Connection.query(
                    method='log.send',
                    param={"msg": msg, "level": level, "subsystem": subsystem, "event_id": event_id, "timestamp": timestamp}
                    )

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            self.Logs.debug(response)

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False


from unrealircd_rpc_py.Definition import RPCError

class EngineError:

    """Engine Error: This attribut will contain the current Error"""

    def __init__(self):
        self.Error = RPCError()
        """Engine Error: This attribut will contain the current Error"""

    def init_error(self):

        self.Error.code = 0
        self.Error.message = ''

    def set_error(self, code:int, message:str):
        """Engine Error set the error

        ### Error Code:
        These are possible Internal Errors:
        ```
        -1: Connection Error
        -2: Empty response
        -3: Exception Error
        ```

        Args:
            code (int): the error code
            message (str): the message of the error
        """
        self.Error.code = code
        self.Error.message = message

class InvalidUrlFormat(Exception):

    def __init__(self, detail: str, returncode: int = -1):
        self.returncode: int = returncode
        detail = detail.strip()
        super().__init__(detail)

class UnixSocketFileNotFoundError(Exception):

    def __init__(self, detail: str, returncode: int = -1):
        self.returncode: int = returncode
        detail = detail.strip()
        super().__init__(detail)


from types import SimpleNamespace
from typing import Union

from unrealircd_rpc_py.Connection import Connection
import unrealircd_rpc_py.Definition as Dfn

class ServerBanException:

    DB_SERVERS_BANS_EXCEPTION: list[Dfn.ServerBanException] = []

    def __init__(self, connection: Connection) -> None:

        # Get the Connection instance
        self.Connection = connection
        self.Logs = connection.Logs
        self.Error = connection.Error

    @property
    def get_error(self) -> Dfn.RPCError:
        return self.Error

    @property
    def get_response(self) -> Union[dict, None]:
        return self.Connection.get_response()

    @property
    def get_response_np(self) -> Union[SimpleNamespace, None]:
        return self.Connection.get_response_np()

    def list_(self) -> list[Dfn.ServerBanException]:
        """List server ban exceptions (ELINEs).

        Returns:
            list[ModelServerBanException]: List of ModelServerBanException, None if nothing see the Error property
        """
        try:

            self.Connection.EngineError.init_error()
            self.DB_SERVERS_BANS_EXCEPTION = []

            response: dict[str, dict] = self.Connection.query(method='server_ban_exception.list')

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return []

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return []

            srvbansexceps = response.get('result', {}).get('list', [])

            for srvbansexcep in srvbansexceps:
                self.DB_SERVERS_BANS_EXCEPTION.append(
                        Dfn.ServerBanException(**srvbansexcep)
                )

            return self.DB_SERVERS_BANS_EXCEPTION

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return []
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return []

    def get(self, name: str) -> Union[Dfn.ServerBanException, None]:
        """Retrieve all details of a single server ban exception (ELINE).

        Mandatory arguments (see structure of a server ban for an explanation of the fields):

        https://www.unrealircd.org/docs/JSON-RPC:Server_ban_exception#Structure_of_a_server_ban_exception

        Args:
            name (str): The target of the ban or except. For Spamfilter this is the regex/matcher.

        Returns:
            ModelServerBanException (ModelServerBanException): The model ModelServerBanException Object | None if nothing see the Error property
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query(method='server_ban_exception.get', param={'name': name})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return None

            srvbanexcep = response['result']['tkl']

            obj = Dfn.ServerBanException(**srvbanexcep)

            return obj

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
        except Exception as err:
            self.Logs.error(f'General error: {err}')

    def add(self, name: str, exception_types: str, reason: str,
            set_by: str = None, expire_at: str = None,
            duration_sting: str = None
            ) -> bool:
        """Add a server ban exception (ELINE).

        Mandatory arguments (see structure of a server ban for an explanation of the fields):

        https://www.unrealircd.org/docs/JSON-RPC:Server_ban_exception#Structure_of_a_server_ban_exception

        Args:
            name (str): user@host mask or extended server ban
            exception_types (str): eg k for a kline exception
            reason (str): The reason of the ban
            expire_at (str, optional): Date/Time when the server ban will expire. NULL means: never. Defaults to None.
            duration_sting (str, optional): How long the ban will last from this point in time (human printable). Uses "permanent" for forever. Defaults to None.
            set_by (str, optional): Name of the person or server who set the ban. Defaults to None.

        Returns:
            bool: True if success
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query(method='server_ban_exception.add', param={"name": name, "exception_types": exception_types, "reason": reason, "expire_at": expire_at, "duration_string": duration_sting, 'set_by': set_by})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False

    def del_(self, name: str, set_by: str = None) -> bool:
        """Delete a server ban exception (ELINE).

        Mandatory arguments (see structure of a server ban for an explanation of the fields):

        https://www.unrealircd.org/docs/JSON-RPC:Server_ban_exception#Structure_of_a_server_ban_exception

        Args:
            name (str): The target of the ban or except. For Spamfilter this is the regex/matcher.
            set_by (str, optional): Name of the person or server who set the ban. Defaults to None.

        Returns:
            bool: True if success
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query(method='server_ban_exception.del', param={"name": name, 'set_by': set_by})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False



from types import SimpleNamespace
from typing import Union
from unrealircd_rpc_py.Connection import Connection
import unrealircd_rpc_py.Definition as Dfn

class Whowas:

    DB_WHOWAS: list[Dfn.Whowas] = []

    def __init__(self, connection: Connection) -> None:

        # Get the Connection instance
        self.Connection = connection
        self.Logs = connection.Logs
        self.Error = connection.Error

    @property
    def get_error(self) -> Dfn.RPCError:
        return self.Error

    @property
    def get_response(self) -> Union[dict, None]:
        return self.Connection.get_response()

    @property
    def get_response_np(self) -> Union[SimpleNamespace, None]:
        return self.Connection.get_response_np()

    def get(self, nick:str = None, ip:str = None, object_detail_level:int = 2) -> list[Dfn.Whowas]:
        """Get WHOWAS history of a user. 6.1.0+

        Args:
            nick (str, optional): The nick name to search for. Defaults to None.
            ip (str, optional): The IP address to search for. Defaults to None.
            object_detail_level (int, optional): set the detail of the response object, this is similar to the Detail level column in Structure of a client object. In this RPC call it defaults to 2 if this parameter is not specified. Defaults to 2.

        Returns:
            List[dfn.Whowas]: The Whowas model or Empty Whowas model
        """
        try:
            self.DB_WHOWAS = []
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query('whowas.get', param={'nick': nick, 'ip': ip, 'object_detail_level': object_detail_level})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return self.DB_WHOWAS

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return self.DB_WHOWAS

            whowass: list[dict] = response.get('result', {}).get('list', []) # ["result"]["list"]

            for whowas in whowass:

                whowas_exclude_objs = whowas.copy()
                whowas_exclude_objs.pop('user', None)
                whowas_exclude_objs.pop('geoip', None)

                self.DB_WHOWAS.append(
                    Dfn.Whowas(
                        **whowas_exclude_objs,
                        user=Dfn.WhowasUser(**whowas.get('user', Dfn.WhowasUser().to_dict())),
                        geoip=Dfn.Geoip(**whowas.get('geoip', Dfn.Geoip().to_dict()))
                    )
                )

            return self.DB_WHOWAS

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return []
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return []



from types import SimpleNamespace
from typing import Union
from unrealircd_rpc_py.Connection import Connection
import unrealircd_rpc_py.Definition as Dfn

class Server:

    DB_SERVER: list[Dfn.ClientServer] = []
    DB_MODULES: list[Dfn.ServerModule] = []

    def __init__(self, connection: Connection) -> None:

        # Get the Connection instance
        self.Connection = connection
        self.Logs = connection.Logs
        self.Error = connection.Error

    @property
    def get_error(self) -> Dfn.RPCError:
        return self.Error

    @property
    def get_response(self) -> Union[dict, None]:
        return self.Connection.get_response()

    @property
    def get_response_np(self) -> Union[SimpleNamespace, None]:
        return self.Connection.get_response_np()

    def list_(self) -> list[Dfn.ClientServer]:
        """List servers.

        Returns:
            list[ClientServer]: List with an object contains all Servers information
        """
        try:
            self.Connection.EngineError.init_error()
            self.DB_SERVER = []

            response: dict[str, dict] = self.Connection.query('server.list')

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return []

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return []

            servers = response['result']['list']

            for server in servers:
                client_server: dict = server.copy()
                server_for_client_server: dict = server.get('server', {}).copy()
                features_for_server: dict = server.get('server', {}).get('features', {}).copy()

                for key in ['server', 'tls']:
                    client_server.pop(key, None)

                for key in ['features']:
                    server_for_client_server.pop(key, None)

                for key in ['rpc_modules']:
                    features_for_server.pop(key, None)

                features_object = Dfn.ServerFeatures(
                    **features_for_server,
                    rpc_modules=[Dfn.ServerRpcModules(**rpcmod) for rpcmod in server.get('server', {}).get('features', {}).get('rpc_modules', [])]
                    )

                server_object = Dfn.Server(
                    **server_for_client_server,
                    features=features_object
                )

                client_server_object = Dfn.ClientServer(
                    **client_server,
                    server=server_object,
                    tls=Dfn.Tls(**server.get('tls', {}))
                )

                self.DB_SERVER.append(client_server_object)

            return self.DB_SERVER

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return []
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return []

    def get(self, serverorsid: str = None) -> Union[Dfn.ClientServer, None]:
        """Retrieve all details of a single server.

        Args:
            serverorsid (str, optional): the server name (or the SID). If not specified then the current server is assumed (the one you connect to via the JSON-RPC API). Defaults to None.

        Returns:
            ClientServer (ClientServer, None): The ClientServer if success | None if error
        """
        try:
            self.Connection.EngineError.init_error()

            response:dict[str, dict] = self.Connection.query('server.get', {'server': serverorsid})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return None

            server = response['result']['server']

            client_server: dict = server.copy()
            server_for_client_server: dict = server.get('server', {}).copy()
            features_for_server: dict = server.get('server', {}).get('features', {}).copy()

            for key in ['server', 'tls']:
                client_server.pop(key, None)

            for key in ['features']:
                server_for_client_server.pop(key, None)

            for key in ['rpc_modules']:
                features_for_server.pop(key, None)

            features_object = Dfn.ServerFeatures(
                **features_for_server,
                rpc_modules=[Dfn.ServerRpcModules(**rpcmod) for rpcmod in server.get('server', {}).get('features', {}).get('rpc_modules', [])]
                )

            server_object = Dfn.Server(
                **server_for_client_server,
                features=features_object
            )

            client_server_object = Dfn.ClientServer(
                **client_server,
                server=server_object,
                tls=Dfn.Tls(**server.get('tls', {}))
            )

            return client_server_object

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return None
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return None

    def rehash(self, serverorsid: str = None) -> Union[Dfn.ServerRehash, None]:
        """Rehash the server.

        IMPORTANT: If all servers on your network have rpc.modules.default.conf included then this can return an object with full rehash details.
        Otherwise, for remote rehashes a simple boolean 'true' result is returned.

        Args:
            serverorsid (str, optional): the server name (or SID). If not specified then the current server is assumed (the one you connect to via the JSON-RPC API). Defaults to None.

        Returns:
            ServerRehash: True if success or False if failed
        """
        try:
            self.Connection.EngineError.init_error()

            response:dict[str, dict] = self.Connection.query('server.rehash', {'server': serverorsid})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return None

            rehash: dict = response['result']

            rehash_log: list[dict] = rehash.get('log', []).copy()

            db_rehash_log: list[Dfn.ServerRehashLog] = []

            for rlog in rehash_log:
                rlog_copy = rlog.copy()
                rlog_copy.pop('source', None)
                db_rehash_log.append(
                    Dfn.ServerRehashLog(
                        **rlog_copy,
                        source=Dfn.ServerRehashLogSource(**rlog.get('source', {}))
                    )
                )

            rehash_object = Dfn.ServerRehash(
                rehash_client=Dfn.ServerRehashClient(**rehash.get("rehash_client", {})),
                log=db_rehash_log,
                success=rehash.get('success', '')
            )

            return rehash_object

        except TypeError as te:
            self.Logs.error(f'Type Error: {te}')
            return None
        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return None
        except Exception as err:
            self.Logs.error(f'General Error: {err}')
            return None

    def connect(self, link: str) -> bool:
        """Make server link (connect) to another server.

        Right now you can only tell to link to servers from the directly connected server 
        (the one you are issuing JSON-RPC calls to), 
        you cannot (yet) tell remote server B to link to server C.

        Args:
            link (str): the server name to link to

        Returns:
            bool: True if success
        """
        try:
            self.Connection.EngineError.init_error()

            response:dict[str, dict] = self.Connection.query('server.connect', {'link': link})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False

    def disconnect(self, link: str) -> bool:
        """Terminate a server link (disconnect).

        This works for both directly connected servers and remote servers further up the network.

        Args:
            link (str): the server name to unlink from

        Returns:
            bool: True if success
        """
        try:
            self.Connection.EngineError.init_error()

            response:dict[str, dict] = self.Connection.query('server.disconnect', {'link': link})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False

    def module_list(self, serverorsid: str = None) -> list[Dfn.ServerModule]:
        """Get the module list (list of loaded modules) on a server.

        IMPORTANT: This only works with remote servers if all servers on your network have rpc.modules.default.conf included.
        This is because then the JSON-RPC request/response is forwarded over the IRC network.

        Args:
            serverorsid (str, optional): The server name (or the SID). If not specified then the current server is assumed (the one you connect to via the JSON-RPC API). Defaults to None.

        Returns:
            ServerModule (ServerModule): if success you will find the object ServerModule
        """
        try:
            self.DB_MODULES = []
            self.Connection.EngineError.init_error()

            response:dict[str, dict] = self.Connection.query('server.module_list', {'server': serverorsid})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return self.DB_MODULES

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return self.DB_MODULES

            modules = response['result']['list']

            for module in modules:
                self.DB_MODULES.append(Dfn.ServerModule(**module))

            return self.DB_MODULES

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return []
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return []



import json
import socket
import os
import base64
import ssl
import re
import time
import logging
import random
import asyncio
from typing import Literal, Union, Optional, Any
from types import SimpleNamespace
from websockets.asyncio import client
from websockets import InvalidURI, InvalidHandshake
from unrealircd_rpc_py.EngineError import EngineError, InvalidUrlFormat, UnixSocketFileNotFoundError
from unrealircd_rpc_py.Definition import RPCError, LiveRPCError, LiveRPCResult

def decompose_url(url: str) -> tuple[str, str, int]:
    """Check provided url if it follow the format
    and decompose it to host, endpoint and port

    >>> # If error it raises InvalidUrlFormat exception
    Args:
        url (str): Url to jsonrpc https://your.rpcjson.link:port/api

    Returns:
        tuple: Host, Endpoint, Port
    """
    if url is None or url == '':
        raise InvalidUrlFormat("The url provided is empty!")

    pattern = r'https?://([a-zA-Z0-9\.-]+):(\d+)/(.+)'

    match = re.match(pattern, url)

    if not match is None:
        host = match.group(1)
        port = match.group(2)
        endpoint = match.group(3)
        return host, endpoint, int(port)
    else:
        raise InvalidUrlFormat("You must provide the url in this format: https://your.rpcjson.link:port/api")

def dict_to_namespace(dictionary: Any) -> Union[SimpleNamespace, Any]:
    """Try to convert a dictionary to a SimpleNamespace object

    :param dictionary:
    :return: SimpleNamespace object or the object if it was impossible to convert
    """
    if isinstance(dictionary, dict):
        return SimpleNamespace(**{key: dict_to_namespace(value) for key, value in dictionary.items()})
    else:
        return dictionary

def verify_unix_socket_file(path_to_socket_file: str) -> bool:
    """Check provided full path to socket file if it exist

    Args:
        path_to_socket_file (str): Full path to unix socket file

    Returns:
        bool: True if path is correct else False
    """
    if path_to_socket_file is None or not os.path.exists(path_to_socket_file):
        raise UnixSocketFileNotFoundError("The socket file is not available, "
                                          "please check the full path of your socket file")

    return True

def remove_logger(logger_name: str) -> None:

    # Get the logger name
    logger = logging.getLogger(logger_name)

    # Delete handlers from the gestionary and close them
    for handler in logger.handlers[:]:  # Use a copy of the list
        # print(handler)
        logger.removeHandler(handler)
        handler.close()

    # Remove the logger from the dictionary
    logging.Logger.manager.loggerDict.pop(logger_name, None)

    return None

def start_log_system(debug_level: int = 20) -> logging.Logger:
    """Init log system
    """
    logger_name = 'unrealircd-liverpc-py'
    remove_logger(logger_name)

    # Init logs object
    logger: logging.Logger = logging.getLogger(logger_name)
    logger.setLevel(debug_level)

    # Add Handlers
    stdout_hanlder = logging.StreamHandler()
    stdout_hanlder.setLevel(debug_level)

    # Define log format
    formatter = logging.Formatter(
        '%(asctime)s - %(levelname)s - %(message)s (%(filename)s::%(funcName)s::%(lineno)d)')

    # Apply log format
    stdout_hanlder.setFormatter(formatter)

    # Add handler to logs
    logger.addHandler(stdout_hanlder)

    return logger

def convert_to_jsonrpc_result(result: Union[dict, Any]) -> LiveRPCResult:

    if not isinstance(result, dict):
        raise TypeError("Invalid jsonrpc response!")

    return LiveRPCResult(**result)

def convert_to_jsonrpc_error(error: Union[dict, Any]) -> LiveRPCError:

    if not isinstance(error, dict):
        raise TypeError("Invalid jsonrpc Error!")

    return LiveRPCError(**error)

class LiveWebsocket:

    def __init__(self,
                 url: str,
                 username: str,
                 password: str,
                 callback_object_instance: object,
                 callback_method_or_function_name: str,
                 debug_level: Literal[10, 20, 30, 40, 50] = 20
                 ):
        """Initiate live connection to unrealircd

            websocket:

            ```
            If you use websocket you must have:
                url: https://your.rpcjson.link:port/api
                username: API_RPC_USERNAME
                password: API_RPC_PASSWORD
            ```

            ## Exemples:
            If you want to use websocket you need to provide 6 parameters:
            ```python
                Live(
                    callback_object_instance=YourCallbackClass,
                    callback_method_name='your_callback_methode_name',
                    url='https://your.rpcjson.link:8600/api',
                    userame='API_RPC_USERNAME',
                    password='API_RPC_PASSWORD'
                )
            ```

            Args:
                callback_object_instance (object): The callback class instance (could be "self" if the class is in the same class)
                callback_method_or_function_name (str): The callback method name
                url (str, optional): The full url to connect https://your.rpcjson.link:port/api.
                username (str, optional): Default to None
                password (str, optional): Default to None
                debug_level (str, optional): NOTSET=0 | DEBUG=10 | INFO=20 | WARN=30 | ERROR=40 | CRITICAL=50 - Default to 20
        """

        self.Logs = start_log_system(debug_level)
        self.EngineError = EngineError()

        self.url = url
        self.username = username
        self.password = password
        self.request: str = ''
        self.connected: bool = True

        self.EngineError.init_error()

        try:
            self.host, self.endpoint, self.port = decompose_url(url)
            self.to_run = getattr(callback_object_instance, callback_method_or_function_name)
            self.Logs.debug("Connexion Established using Live Websocket!")

        except AttributeError as atterr:
            self.EngineError.set_error(code=-1, message=f'CallbackMethodError: {atterr}')
            self.Logs.error(f'CallbackMehtodError: {atterr}')
            return

        except InvalidUrlFormat as iuf:
            self.EngineError.set_error(code=-1, message=iuf.__str__())
            self.Logs.critical(iuf)
            return

    # No more asyncio.run
    async def subscribe(self, sources: Optional[list] = None) -> dict[str, Any]:
        """Subscribe to the rpc server stream"""
        self.connected = True
        sources = ["!debug", "all"] if sources is None else sources
        response = await self.query('log.subscribe', param={"sources": sources})
        return response

    async def unsubscribe(self) -> dict[str, Any]:
        """Unsubscribe from the rpc server stream"""

        response = await self.query(method='log.unsubscribe')

        await self.query(method='log.send',
                         param={"msg": f"{self.username} JSONRPC User has been disconnected from the stream!",
                                "level": "info",
                                "subsystem": "connect",
                                "event_id": "REMOTE_CLIENT_DISCONNECT"}
                         )

        return response

    async def query(self, method: Union[Literal['log.subscribe', 'log.unsubscribe'], str],
                    param: Optional[dict] = None, query_id: int = 123,
                    jsonrpc: str = '2.0'
                    ) -> dict[str, Any]:

        get_param = {} if param is None else param
        get_id = int(time.time()) + random.randint(1, 6000) if query_id == 123 else query_id

        response = {
            "jsonrpc": jsonrpc,
            "method": method,
            "params": get_param,
            "id": get_id
        }

        self.request = json.dumps(response)

        response = await self.__send_websocket()

        return response

    async def __send_websocket(self) -> dict[str, Any]:
        """Connect using websockets"""
        try:
            api_login = f'{self.username}:{self.password}'
            credentials = base64.b64encode(api_login.encode()).decode()

            headers = {'Authorization': f"Basic {credentials}"}

            sslctx = ssl.create_default_context()
            sslctx.check_hostname = False
            sslctx.verify_mode = ssl.CERT_NONE

            ws_uri = f'wss://{self.host}:{self.port}/'
            method = json.loads(self.request).get('method')
            final_response: Optional[Union[LiveRPCResult, LiveRPCError]] = LiveRPCResult()

            async with client.connect(uri=ws_uri, additional_headers=headers, ssl=sslctx) as ws:
                await ws.send(self.request)
                while self.connected:
                    srv_response = await ws.recv()
                    decoded_response: dict[str, Any] = json.loads(json.loads(json.dumps(srv_response)))

                    if decoded_response.get('result', None):
                        final_response = convert_to_jsonrpc_result(decoded_response)

                    if decoded_response.get('error', None):
                        final_response = convert_to_jsonrpc_error(decoded_response)

                    if method == 'log.unsubscribe':
                        self.connected = False
                        final_response = LiveRPCError(error=RPCError(code=0, message="Websocket Normal Closure!"))

                    # support callbacks async et sync
                    result = self.to_run(dict_to_namespace(final_response.to_dict()))
                    if asyncio.iscoroutine(result):
                        await result

                    self.EngineError.init_error()

            return final_response.to_dict()

        except asyncio.CancelledError:
            self.Logs.info("Websocket task cancelled, closing connection")
            self.EngineError.set_error(code=-1, message='Websocket task cancelled!')
            error = LiveRPCError(error=RPCError(code=-1, message="Websocket task cancelled, closing connection")).to_dict()
            result = self.to_run(dict_to_namespace(error))
            if asyncio.iscoroutine(result):
                await result

            return LiveRPCError(error=RPCError(code=-1, message="Websocket task cancelled, closing connection")).to_dict()

        except AttributeError as ae:
            self.Logs.critical(f"Attribute Error: {ae.__str__()} Check your callback function or method")
            self.EngineError.set_error(code=-1, message='Attribute Error, Check your callback function or method!')
            return LiveRPCError(error=RPCError(code=-1, message=ae.__str__())).to_dict()

        except (KeyError, InvalidURI, InvalidHandshake, json.decoder.JSONDecodeError,
                TimeoutError, OSError, TypeError, Exception) as err:
            self.Logs.error(f"Websocket Error: {err}")
            self.Logs.error(f"Initial request: {self.request}")
            self.EngineError.set_error(code=-1, message=f'Websocket Error {err}')

            error = LiveRPCError(error=RPCError(code=-1, message=err)).to_dict()
            result = self.to_run(dict_to_namespace(error))
            if asyncio.iscoroutine(result):
                await result

            return error

    @property
    def get_error(self) -> RPCError:
        return self.EngineError.Error

class LiveUnixSocket:

    def __init__(self,
                 callback_object_instance: object,
                 callback_method_or_function_name: str,
                 path_to_socket_file: str = None,
                 debug_level: Literal[10, 20, 30, 40, 50] = 20
                 ):
        """Initiate live connection to unrealircd

        unixsocket:
        ```
        If you use unixsocket as req_method you must provide:
            path_to_socket_file
        ```

        ## Exemples:
        If you want to use unixsocket you need to provide 4 parameters:
        ```python
            Live(
                req_method='unixsocket',
                callback_object_instance=YourCallbackClass,
                callback_method_or_function_name='your_callback_methode_name',
                path_to_socket_file='/path/to/unrealircd/socket/your_socket.socket'
            )
        ```

        Args:
            callback_object_instance (object): The callback class instance (could be "self" if the class is in the same class)
            callback_method_or_function_name (str): The callback method name
            path_to_socket_file (str, optional): The full path to your unix socket file
            debug_level (str, optional): NOTSET=0 | DEBUG=10 | INFO=20 | WARN=30 | ERROR=40 | CRITICAL=50 - Default to 20
        """

        self.Logs = start_log_system(debug_level)
        """Starting logger system"""

        self.EngineError = EngineError()
        """Engine Error: should be used to set errors"""

        self.request: str = ''
        self.connected: bool = True
        self.__response: Optional[dict] = {}
        self.__response_np: Optional[SimpleNamespace] = SimpleNamespace()

        try:
            if verify_unix_socket_file(path_to_socket_file):
                self.path_to_socket_file = path_to_socket_file
                self.to_run = getattr(callback_object_instance, callback_method_or_function_name)
                self.Logs.debug("Connexion Established using Live UnixSocket!")

        except AttributeError as atterr:
            self.EngineError.set_error(code=-1, message=f'CallbackMethodError: {atterr}')
            self.Logs.error(f'CallbackMehtodError: {atterr}')
            return

        except UnixSocketFileNotFoundError as err:
            self.EngineError.set_error(code=-1, message=err.__str__())
            self.Logs.critical(f'UnixSocketFileNotFoundError: {err}')
            return


    async def subscribe(self, sources: Optional[list] = None) -> dict[str, Any]:
        """Subscribe to the rpc server stream
        sources exemple:
        \n ["!debug","all"] would give you all log messages except for debug messages
        see: https://www.unrealircd.org/docs/List_of_all_log_messages

        Args:
            sources (list, optional): The ressources you want to subscribe. Defaults to ["!debug","all"].
        """
        self.connected = True
        sources = ["!debug", "all"] if sources is None else sources
        response = await self.query(method='log.subscribe', param={"sources": sources})
        return response

    async def unsubscribe(self) -> dict[str, Any]:
        """Run a del timer to trigger an event and then unsubscribe from the stream
        """
        response = await self.query(method='log.unsubscribe')
        await self.query(method='log.send',
                         param={"msg": f"JSONRPC UnixSocket has been disconnected from the stream!",
                                "level": "info",
                                "subsystem": "connect",
                                "event_id": "REMOTE_CLIENT_DISCONNECT"}
                         )

        return response

    async def query(self, method: Union[Literal['log.subscribe', 'log.unsubscribe'], str],
                    param: Optional[dict] = None, query_id: int = 123,
                    jsonrpc: str = '2.0'
                    ) -> dict[str, Any]:
        """This method will use to run the queries

        Args:
            method (Union[Literal[&#39;stats.get&#39;, &#39;rpc.info&#39;,&#39;user.list&#39;], str]): The method to send to unrealircd
            param (dict, optional): the paramaters to send to unrealircd. Defaults to {}.
            query_id (int, optional): id of the request. Defaults to 123.
            jsonrpc (str, optional): jsonrpc. Defaults to '2.0'.

        Returns:
            str: The correct response
            None: no response from the server
            bool: False if there is an error occured
        """

        # data = '{"jsonrpc": "2.0", "method": "stats.get", "params": {}, "id": 123}'
        get_method = method
        get_param = {} if param is None else param

        if query_id == 123:
            rand = random.randint(1, 6000)
            get_id = int(time.time()) + rand
        else:
            get_id = query_id

        response = {
            "jsonrpc": jsonrpc,
            "method": get_method,
            "params": get_param,
            "id": get_id
        }

        self.request = json.dumps(response)

        response = await self.__send_to_permanent_unixsocket()

        return response

    async def __send_to_permanent_unixsocket(self) -> dict[str, Any]:
        sock = socket.socket(socket.AddressFamily.AF_UNIX, socket.SocketKind.SOCK_STREAM)
        try:
            sock.connect(self.path_to_socket_file)

            if not self.request:
                return {}

            # Sending the request
            sock.sendall(f'{self.request}\r\n'.encode())

            # Get the method
            method = json.loads(self.request).get('method')

            # Init batch variable
            batch = b''
            final_response: Optional[Union[LiveRPCResult, LiveRPCError]] = LiveRPCResult()

            while self.connected:
                # Recieve the data from the rpc server, decode it and split it
                response = sock.recv(4096)
                if response[-1:] != b"\n":
                    # If END not recieved then fill the batch and go to next itteration
                    batch += response
                    continue
                else:
                    # The EOF is found, include current response to the batch and continue the code
                    batch += response

                # Decode and split the response
                response = batch.decode().split("\n")

                if method == 'log.unsubscribe':
                    self.connected = False
                    final_response = LiveRPCError(error=RPCError(code=0, message="Websocket Normal Closure!"))

                    # support callbacks async et sync
                    result = self.to_run(dict_to_namespace(final_response.to_dict()))
                    if asyncio.iscoroutine(result):
                        await result

                    break

                for bdata in response:
                    if bdata:
                        decoded_response = json.loads(bdata)

                        if decoded_response.get('result', None):
                            final_response = convert_to_jsonrpc_result(decoded_response)
                        elif decoded_response.get('error', None):
                            final_response = convert_to_jsonrpc_error(decoded_response)

                        # support callbacks async et sync
                        result = self.to_run(dict_to_namespace(final_response.to_dict()))
                        if asyncio.iscoroutine(result):
                            await result

                # Clean batch variable
                batch = b''

                # Clean error
                self.EngineError.init_error()

            return final_response.to_dict()

        except AttributeError as attrerr:
            self.Logs.critical(f'AF_Unix Error: {attrerr}')
            self.EngineError.set_error(code=-1, message=f'AF_UnixError: {attrerr}')

            error = LiveRPCError(error=RPCError(code=-1, message=attrerr.__str__())).to_dict()
            result = self.to_run(dict_to_namespace(error))
            if asyncio.iscoroutine(result):
                await result

            return error
        except (TimeoutError, OSError, json.decoder.JSONDecodeError, TypeError, Exception) as err:
            self.Logs.critical(f'UnixSocket Error: {err}')
            self.EngineError.set_error(code=-1, message=f'UnixSocket Error: {err}')

            error = LiveRPCError(error=RPCError(code=-1, message=err)).to_dict()
            result = self.to_run(dict_to_namespace(error))
            if asyncio.iscoroutine(result):
                await result

            return error
        finally:
            sock.close()

    @property
    def get_error(self) -> RPCError:
        return self.EngineError.Error



from types import SimpleNamespace
from typing import Union

from unrealircd_rpc_py.Connection import Connection
import unrealircd_rpc_py.Definition as Dfn

class NameBan:

    DB_NAME_BANS: list[Dfn.NameBan] = []

    def __init__(self, connection: Connection) -> None:

        # Get the Connection instance
        self.Connection = connection
        self.Logs = connection.Logs
        self.Error = connection.Error

    @property
    def get_error(self) -> Dfn.RPCError:
        return self.Error

    @property
    def get_response(self) -> Union[dict, None]:
        return self.Connection.get_response()

    @property
    def get_response_np(self) -> Union[SimpleNamespace, None]:
        return self.Connection.get_response_np()

    def list_(self) -> list[Dfn.NameBan]:
        """List name bans (qlines).

        Returns:
            ModelNameBan: List of ModelNameBan, None if nothing see Error property
        """
        try:
            self.DB_NAME_BANS = []
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query(method='name_ban.list')

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return []

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return []

            namebans: list[dict] = response.get('result', {}).get('list', []) # ['result']['list']

            for nameban in namebans:
                self.DB_NAME_BANS.append(
                        Dfn.NameBan(**nameban)
                )

            return self.DB_NAME_BANS

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return []
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return []

    def get(self, name: str) -> Union[Dfn.NameBan, None]:
        """Retrieve all details of a single name ban (qline).

        Args:
            name (str): name of the ban, eg *C*h*a*n*s*e*r*v*

        Returns:
            Union[ModelNameBan, None, bool]: The Object ModelNameBan, None if nothing see Error property
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query(method='name_ban.get', param={"name": name})

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return None

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return None

            nameban = response['result']['tkl']

            obj = Dfn.NameBan(**nameban)

            return obj

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return None
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return None

    def add(self, name: str, reason: str, set_by: str = None, expire_at: str = None, duration_string: str = None) -> bool:
        """Add a name ban (qline).

        Mandatory arguments (see structure of a name ban for an explanation of the fields):

        https://www.unrealircd.org/docs/JSON-RPC:Name_ban#Structure_of_a_name_ban

        Args:
            name (str): The target of the ban or except. For Spamfilter this is the regex/matcher.
            reason (str): The reason of the ban
            set_by (str, optional): Name of the person or server who set the ban. Default to None
            expire_at (str, optional): Date/Time when the server ban will expire. NULL means: never. Default to None
            duration_string (str, optional): How long the ban will last from this point in time (human printable). Uses "permanent" for forever. Default to None

        Returns:
            bool: True if success
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query(
                method='name_ban.add', 
                param={"name": name, "reason": reason, "set_by": set_by, "expire_at": expire_at, "duration_string": duration_string}
                )

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False

    def del_(self, name: str, set_by: str = None) -> bool:
        """Delete a name ban (LINE).

        Args:
            name (str): The target of the ban or except. For Spamfilter this is the regex/matcher.
            set_by (str, optional): Name of the person or server who set the ban. Default to None

        Returns:
            bool: True if success
        """
        try:
            self.Connection.EngineError.init_error()

            response: dict[str, dict] = self.Connection.query(
                method='name_ban.del',
                param={"name": name, 'set_by': set_by}
                )

            if response is None:
                self.Logs.error('Empty response')
                self.Connection.EngineError.set_error(code=-2, message='Empty response')
                return False

            if 'error' in response:
                self.Logs.error(response['error']['message'])
                self.Connection.EngineError.set_error(**response["error"])
                return False

            if 'result' in response:
                if response['result']:
                    self.Logs.debug(response)
                    return True

            return True

        except KeyError as ke:
            self.Logs.error(f'KeyError: {ke}')
            return False
        except Exception as err:
            self.Logs.error(f'General error: {err}')
            return False
